void module_clear(void)
{
	memset(&module,0,sizeof(moduleStruct));

	curRow=0;
	curColumn=0;
	curSelectionActive=false;
	curInstrument=0;
	curAutoStepCount=0;
	curResetValue=true;
	activeTrack = 0;

	module.loopStart=0;
	module.loopEnd=MAX_ROWS-1;
	module.speed=6;
	module.speed_min = 1;
	module.speed_max = 99;
	module.speed_column = -1;	//no speed column by default
	module.measure=4;

	strncpy(module.title, "untitled", sizeof(module.title));
	strncpy(module.author, "uncredited", sizeof(module.author));

	update_selection();

	for (int col = 0; col < MAX_COLUMNS; ++col)
	{
		channelMute[0][col] = false;
		channelMute[1][col] = false;
	}

	undoAvailable=false;

	change_engine();

	update_window_title();
}



bool module_is_row_empty(int track, int row, bool marker)
{
	if(marker) if (module.marker[row]) return false;

	for (int col = 0; col < module.columns; ++col)
	{
		if (module.columnType[col] != COLUMN_TYPE_EMPTY)
		{
			if (module.rows[track][row].column[col])
			{
				return false;
			}
		}
	}

	return true;
}



int module_find_last_empty_row(int track)
{
	for (int row = MAX_ROWS - 1; row >= 0; --row)
	{
		if (!module_is_row_empty(track, row, true)) return row + 1;
	}

	return 1;
}



int instrument_get_actual_length(int instrument)
{
	for (int i = MAX_INSTRUMENT_DATA - 1; i >= 0; --i)
	{
		if (module.instrument[instrument].data[i]) return i + 1;
	}

	return 0;
}



bool load_save_error(const char *str, const char *name)
{
	snprintf(engineErrorMessage, sizeof(engineErrorMessage), "%s '%s'", str, name);

	engineErrorMessageActive = true;

	return false;
}



void module_save_track(FILE* file, int track)
{
	int rows = module_find_last_empty_row(track);

	for (int row = 0; row < rows; ++row)
	{
		if (module_is_row_empty(track, row, true)) continue;
			
		xfprintf(file, "%4.4i%c", row, module.marker[row] ? '*' : ' ');

		for (int col = 0; col < module.columns; ++col)
		{
			int ct = module.columnType[col];
			int val = module.rows[track][row].column[col];

			switch (ct)
			{
			case COLUMN_TYPE_EMPTY:
				xfprintf(file, " ");
				break;

			case COLUMN_TYPE_LETTER:
				xfprintf(file, "%c", val ? val : '.');
				break;

			case COLUMN_TYPE_NOTE:
				xfprintf(file, "%s", noteNames[val]);
				break;

			default:
				for (int k = 0; k < ct; ++k) xfprintf(file, "%c", NUM2DEC(val, ct - 1 - k));
			}
		}

		if (!row || module.marker[row])
		{
			if (module.markerName[row]) xfprintf(file, "%s", module.markerName[row]);
		}

		xfprintf(file, "\n");
	}
}



bool module_save(const char* filename)
{
	FILE* file = xfopen(filename, "wt");

	if (!file) return load_save_error("Can't save module", filename);

	//signature

	xfprintf(file,"%s\n\n",SONG_SIGNATURE);

	//engine

	xfprintf(file,"Engine=%s\n\n",engineFileName);

	//song settings

	xfprintf(file,"[Settings]\n\n");

	xfprintf(file,"Title=%s\n" ,module.title);
	xfprintf(file,"Author=%s\n",module.author);
	xfprintf(file,"Speed=%i\n",module.speed);
	xfprintf(file,"LoopStart=%i\n",module.loopStart);
	xfprintf(file,"LoopEnd=%i\n",module.loopEnd);
	xfprintf(file,"Measure=%i\n",module.measure);

	//pattern format, to keep it human readible in the text, and be able to easily parse back

	xfprintf(file,"PatternFormat=");

	for (int col = 0; col < module.columns; ++col)
	{
		switch(module.columnType[col])
		{
		case COLUMN_TYPE_EMPTY:  xfprintf(file, "_"); break;
		case COLUMN_TYPE_LETTER: xfprintf(file, "e"); break;
		case COLUMN_TYPE_NOTE: xfprintf(file, "N"); break;
		default: xfprintf(file, "%i", module.columnType[col]);
		}
	}

	xfprintf(file,"\n\n");

	//instruments

	xfprintf(file,"[Instruments]\n\n");

	for (int i = 0; i < module.instruments; ++i)
	{
		int len = instrument_get_actual_length(i);

		if(len)
		{
			xfprintf(file, "Instrument%i=", i + 1);

			for (int j = 0; j < len; ++j) xfprintf(file, "%2.2x", module.instrument[i].data[j]);

			xfprintf(file, "\n");
		}
	}

	xfprintf(file,"\n");

	//song text

	xfprintf(file, "[Song]\n\n");

	module_save_track(file, 0);

	xfprintf(file, "\n[Side]\n\n");

	module_save_track(file, 1);

	xfclose(file);

	update_window_title();
	return true;
}



int module_find_tag(char* data,int size,const char* tag)
{
	int len = strlen(tag);
	int ptr = 0;

	while(size)
	{
		if (!memcmp(&data[ptr], tag, len)) return ptr + len;

		++ptr;
		--size;
	}

	return -1;
}



char* module_load_str(char* data,int size,const char* tag)
{
	static char str[1024];

	int len=strlen(tag);
	int ptr=0;

	while(size)
	{
		if(!memcmp(data,tag,len))
		{
			data+=len;

			while(size)
			{
				char c = *data++;
				if(c<32) break;

				str[ptr++]=c;
			}

			str[ptr]=0;

			return str;
		}

		++data;
		--size;
	}

	return NULL;
}



int module_load_int(char* data,int size,const char* tag)
{
	int len=strlen(tag);

	while(size)
	{
		if(!memcmp(data,tag,len))
		{
			data+=len;

			int n=0;

			while(size)
			{
				char c=*data++;
				if(c<'0'||c>'9') break;

				n=n*10+(c-'0');
				--size;
			}

			return n;
		}

		++data;
		--size;
	}

	return -1;
}



bool is_num(char c)
{
	return ((c >= '0') && (c <= '9')) ? true : false;
}



void module_load_track(char* text, int text_size, int track, const char* section)
{
	char markerName[MAX_MARKER_NAME + 1];

	int ptr = module_find_tag(text, text_size, section);

	while (ptr < text_size)
	{
		while (ptr < text_size) if (text[ptr] < 32) ++ptr; else break;

		if (ptr >= text_size) break;
		if (text[ptr] == '[') break;

		bool no_row = false;
		int row = 0;

		if (is_num(text[ptr + 0])) row += (text[ptr + 0] - '0') * 1000; else no_row = true;
		if (is_num(text[ptr + 1])) row += (text[ptr + 1] - '0') * 100; else no_row = true;
		if (is_num(text[ptr + 2])) row += (text[ptr + 2] - '0') * 10; else no_row = true;
		if (is_num(text[ptr + 3])) row += (text[ptr + 3] - '0') * 1; else no_row = true;

		ptr += 4;

		if (no_row) break;

		if (text[ptr++] == '*') module.marker[row] = true;

		int ptr_max = ptr;

		while (ptr_max < text_size)
		{
			if (text[ptr_max] < 32) break;
			++ptr_max;
		}

		for (int col = 0; col < module.columns; ++col)
		{
			int ct = module.columnType[col];

			switch (ct)
			{
			case COLUMN_TYPE_EMPTY:
			{
				if (ptr < ptr_max) ++ptr;
			}
			break;

			case COLUMN_TYPE_LETTER:
			{
				if (ptr < ptr_max)
				{
					int val = text[ptr++];
					module.rows[track][row].column[col] = val != '.' ? val : 0;
				}
			}
			break;

			case COLUMN_TYPE_NOTE:
			{
				if (ptr < ptr_max)
				{
					for (int j = 0; j < noteNamesCount; ++j)
					{
						if (!memcmp(&text[ptr], noteNames[j], 3))
						{
							module.rows[track][row].column[col] = j;
							break;
						}
					}

					ptr += 3;
				}
			}
			break;

			default:
			{
				if (ptr < ptr_max)
				{
					int n = 0;

					for (int j = 0; j < ct; ++j)
					{
						int m = text[ptr++];

						if (m == '.') m = 0; else m = m - '0';

						n = n * 10 + m;
					}

					module.rows[track][row].column[col] = n;
				}
			}
			}
		}

		//read optional marker name

		while (text[ptr] == 32) ++ptr;

		int n = 0;

		while (text[ptr] >= 32)
		{
			if (n < MAX_MARKER_NAME) markerName[n++] = text[ptr];

			++ptr;
		}

		if (n)
		{
			markerName[n] = 0;

			if (!row || module.marker[row])
			{
				strncpy(module.markerName[row], markerName, sizeof(module.markerName[row]));
			}
		}
	}
}



bool module_load(const char* filename)
{
	char insName[1024];

	FILE* file = xfopen(filename, "rb");

	if(!file) return load_save_error("Can't load module", filename);

	xfseek(file,0,SEEK_END);
	int size=xftell(file);
	xfseek(file,0,SEEK_SET);

	char* text=(char*)malloc(size+1);
	text[size-1]=0;

	xfread(text,size,1,file);
	xfclose(file);

	//check signature

	if(module_find_tag(text,size,SONG_SIGNATURE)<0)
	{
		free(text);
		return false;
	}

	//song settings

	strncpy(engineFileName, module_load_str(text, size, "Engine="), sizeof(engineFileName));

	change_engine();
	module_clear();

	module.speed    =module_load_int(text,size,"Speed=");
	module.loopStart=module_load_int(text,size,"LoopStart=");
	module.loopEnd  =module_load_int(text,size,"LoopEnd=");
	module.measure  =module_load_int(text,size,"Measure=");

	strncpy(module.title, module_load_str(text, size, "Title="), sizeof(module.title));
	strncpy(module.author, module_load_str(text, size, "Author="), sizeof(module.author));

	//instruments

	for (int ins = 0; ins < module.instruments; ++ins)
	{
		snprintf(insName, sizeof(insName), "Instrument%i=", ins + 1);

		int ptr = module_find_tag(text, size, insName);

		if(ptr>=0)
		{
			int insSize=0;
			int nibble=0;
			int m = 0;

			while(ptr<size)
			{
				int n=text[ptr++];

				if(n<32) break;

				if (!nibble) m = HEX2DEC(n) << 4; else module.instrument[ins].data[insSize++] = m | HEX2DEC(n);

				nibble^=1;
			}

			module.instrument[ins].data[insSize] = 0;	//for null-terminated text strings in the instrument data
		}
	}

	//song text

	module_load_track(text, size, 0, "[Song]");
	module_load_track(text, size, 1, "[Side]");

	free(text);

	return true;
}



bool instrument_save(const char* filename,int ins)
{
	FILE *file;
	int len,i;

	file=xfopen(filename,"wt");

	if (!file) return load_save_error("Can't save instrument", filename);

	//signature

	xfprintf(file,"%s\n\n",INSTRUMENT_SIGNATURE);

	//instruments

	len=instrument_get_actual_length(ins);

	if(len)
	{
		xfprintf(file,"InstrumentData=");

		for(i=0;i<len;++i) xfprintf(file,"%2.2x",module.instrument[ins].data[i]);

		xfprintf(file,"\n");
	}

	xfprintf(file,"\n");

	xfclose(file);
	
	return true;
}



bool instrument_load(const char* filename,int ins)
{
	FILE *file;
	char *text;
	int n,m,ptr,size,insSize,nibble;

	file=xfopen(filename,"rb");

	if(!file) return load_save_error("Can't load instrument", filename);

	xfseek(file,0,SEEK_END);
	size=xftell(file);
	xfseek(file,0,SEEK_SET);

	text=(char*)malloc(size+1);
	text[size-1]=0;

	xfread(text,size,1,file);
	xfclose(file);

	//check signature

	if(module_find_tag(text,size,INSTRUMENT_SIGNATURE)<0)
	{
		free(text);
		return false;
	}

	ptr=module_find_tag(text,size,"InstrumentData=");

	if(ptr>=0)
	{
		memset(module.instrument[ins].data,0,MAX_INSTRUMENT_DATA);

		insSize=0;
		nibble=0;
		m = 0;

		while(ptr<size)
		{
			n=text[ptr++];

			if(n<32) break;

			if(!nibble) m=HEX2DEC(n)<<4; else module.instrument[ins].data[insSize++]=m|HEX2DEC(n);

			nibble^=1;
		}
	}
	else
	{
		return false;
	}

	free(text);

	return true;
}



void do_backup(void)
{
	char backupName[MAXPATH];
	time_t rawTime;
	struct tm *timeInfo;

	if (!backupPeriod) return;

#ifdef WIN32
	_mkdir(backupDirectory);
#else
	mkdir(backupDirectory, S_IRWXU);
#endif

	time(&rawTime);
	timeInfo = localtime(&rawTime);

	snprintf(backupName, sizeof(backupName), "%s%2.2i%2.2i%2.2i_%2.2i%2.2i.1tm", backupDirectory, timeInfo->tm_mday, timeInfo->tm_mon, timeInfo->tm_year % 100, timeInfo->tm_hour, timeInfo->tm_min);

	module_save(backupName);
}



int GetSongSpeed(); 
int GetSongSpeedMin();
int GetSongSpeedMax();
int GetSongLength();
void delete_row_with_shift(int);

void module_optimize_row_speed(void)
{
	if (module.speed_column < 0) return;

	set_undo();
	
	while (1)
	{
		int speed = GetSongSpeed();
		int row = 0;
		int len = GetSongLength() - 1;

		if (!module.rows[activeTrack][row].column[module.speed_column])
		{
			module.rows[activeTrack][row].column[module.speed_column] = speed;
		}

		bool change = false;
		
		while (row < len)
		{
			if (module.rows[activeTrack][row].column[module.speed_column] > 0)	//speed column
			{
				speed = module.rows[activeTrack][row].column[module.speed_column];
			}

			if (!module_is_row_empty(activeTrack, row, true) && module_is_row_empty(activeTrack, row + 1, true))
			{
				if (speed * 2 <= GetSongSpeedMax())
				{
					if (row + 2 < MAX_ROWS)
					{
						module.rows[activeTrack][row].column[module.speed_column] = speed * 2;
						delete_row_with_shift(row + 2);
						--len;

						if (module.rows[activeTrack][row + 1].column[module.speed_column] == 0) module.rows[activeTrack][row + 1].column[module.speed_column] = speed;
						change = true;
					}
				}
			}
			
			++row;
		}

		//cleanup repeating speed change values

		speed = GetSongSpeed();

		for (int row = 0; row < len; ++row)
		{
			int row_speed = module.rows[activeTrack][row].column[module.speed_column];	//speed column

			if (row_speed > 0)
			{
				if (row_speed == speed)
				{
					module.rows[activeTrack][row].column[module.speed_column] = 0;
				}
				else
				{
					speed = row_speed;
				}
			}
		}

		if (!change) break;
	}
}